

// Computes UBRR with the least error for the given baud rate and cpu
// frequency. The factor is 16 for normal speed and 8 for double speed mode. 
// The return value is the error in Hz for the conditions. Calculated
// UBRR value is returned in *divider.
uint32_t delta_and_divider(uint32_t baud, uint32_t factor, uint32_t *divider)
{
  uint32_t D, F1, F2;
  *divider= D= F_CPU/factor/baud - 1;

  F1= (D+1)*baud*factor;
  F2= (D+2)*baud*factor;
  if (F2-F_CPU<F_CPU-F1) { (*divider)++; return F2-F_CPU; }
                   else  { return F_CPU-F1; }
}

// This function finds the best combination of UBRR and speed mode
// for the given baud rate. The return value is 1 if double speed
// is to be used, 0 otherwise. The UBRR value is stored in *divider.
uint8_t compute_divider(uint32_t baud, uint32_t *divider)
{
  uint32_t divider_n, divider_d;

  if (delta_and_divider(baud, 8, &divider_d) < 
      delta_and_divider(baud,16, &divider_n))
    { *divider= divider_d; return 1; }
  else
    { *divider= divider_n; return 0; }
}

void usart_set_rate(uint32_t baud)
{
  uint32_t divider;

  if (compute_divider(baud, &divider))
    { lcd_char('D'); UCSR0A= (1<<U2X0); }
  else
    { lcd_char('N'); UCSR0A= 0; }

  UBRR0H = (divider >> 8 ) & 0x0f;
  UBRR0L = (divider & 0xff);

  phex(divider>>8);
  phex(divider&0xff);

  UCSR0C= (1<<UCSZ01) | (1<<UCSZ00);
  UCSR0B= (1<<RXCIE0) | (1<<RXEN0);
}

ISR(USART_RX_vect)
{
  uint8_t D;
  uint16_t K;
  D= UDR0;
  if (waitchar && D==waitchar) 
   {
     waitchar= 0; 
     // b_rx++; 
     return ; 
   }
  K= (rxtail+1)%512;
  if (K!=rxhead) 
  {
    rxbuf[rxtail]= D;
    rxtail= K;
  }
  // b_rx++;
}

void usart_send(uint8_t *buf, int len)
{
  if (!len) return ;
  int i;
  UCSR0B |= (1<<TXEN0);
  for(i=0;i<len;i++)
  {
     while(!(UCSR0A&(1<<UDRE0))) ;
     UDR0= buf[i];
     b_tx++;
  }
  while(!(UCSR0A&(1<<UDRE0))) ;
  UCSR0B &= ~(1<<TXEN0);
}

void usart_send_start()
{
  UCSR0B |= (1<<TXEN0);
}

void usart_send_stop()
{
  while(!(UCSR0A&(1<<UDRE0))) ;
  UCSR0B &= ~(1<<TXEN0);
}

void usart_send_pure(uint8_t V)
{
  while(!(UCSR0A&(1<<UDRE0)));
  UDR0= V;
  b_tx++;
}

void usart_stop()
{
  UCSR0B &= ~((1<<TXEN0) | (1<<RXEN0));
}

void usart_send_str(char *k)
{
  usart_send((uint8_t*) k,strlen(k));
}

void usart_send_chr(char z)
{
  usart_send((uint8_t*) &z, 1);
}


void usart_send_str_raw(char *k)
{
  int i;
  for(i=0;k[i];i++) usart_send_pure(k[i]);
}

void usart_send_progmem(const char *str)
{
  uint8_t K;
  while(1)
  {
    K= pgm_read_byte(str);
    if (!K) break;
    usart_send_pure(K);
    str++;
  }
}

void usart_send_decimal(uint32_t V)
{
  uint32_t D;
  uint8_t Z,f;
  f= 0;
  for(D=1000000UL;D!=0;D/=10)
  {
    Z= 0;
    for(Z=0;V>=D;Z++) V-= D;
    if (D==1 || Z || f) { usart_send_pure(Z+'0'); f= 1; }
  }
}
